/* glwin.cpp - 03 Apr 09
   Hosts3D - 3D Real-Time Network Monitor
   Copyright (c) 2006-2009  Del Castle

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details. */

#include <ctype.h>  //tolower()
#include <stdio.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "glwin.h"

//object type
#define GLWIN_WINDOW  0x01
#define GLWIN_BUTTON  0x02
#define GLWIN_INPUT   0x03
#define GLWIN_LABEL   0x04
#define GLWIN_LIST    0x05
#define GLWIN_VIEW    0x06

const unsigned char MTF[] = "...";  //more to follow

MyGLWin::MyGLWin()
{
  Reset();
}

void MyGLWin::Reset()
{
  lastWin = 0;
  lastInput = 0;
  names = 1;
  selection = 0;
  currInput = 0;
  currScroll = 0;
}

//draw window object
void MyGLWin::DrawWin(GLenum mode, glwn_obj *wn)
{
  int pt = glutGet(GLUT_WINDOW_HEIGHT) - wn->top, pr = wn->left + wn->width, pb = pt - wn->height;
  if (mode == GL_RENDER)
  {
    //glEnable(GL_BLEND);
    //glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    glColor3ub(0, 0, 50);
    glRecti(wn->left, pb, pr, pt);
    //glDisable(GL_BLEND);
    glColor3ub(0, 0, 255);
    glBegin(GL_LINE_STRIP);
    glVertex2i(wn->left, pt - 20);
    glVertex2i(wn->left, pb);
    glVertex2i(pr, pb);
    glVertex2i(pr, pt - 20);
    glVertex2i(wn->left, pt - 20);
    glVertex2i(wn->left, pt);
    glVertex2i(pr, pt);
    glVertex2i(pr, pt - 20);
    glEnd();
    glColor3ub(255, 255, 255);
    glRasterPos2i(wn->left + 10, pt - 14);
    glutBitmapString(GLUT_BITMAP_HELVETICA_10, (const unsigned char *)wn->title);  //title bar text
  }
  else  //title bar selectable
  {
    glLoadName((wn->name * 10000) + GLWIN_TITLE);
    glRecti(wn->left, pt - 20, pr - 20, pt);
  }
  if (wn->close)  //close object
  {
    if (mode == GL_RENDER)
    {
      glColor3ub(0, 0, 255);
      glBegin(GL_LINE_LOOP);
      glVertex2i(pr - 16, pt - 16);
      glVertex2i(pr - 4, pt - 16);
      glVertex2i(pr - 4, pt - 4);
      glVertex2i(pr - 16, pt - 4);
      glEnd();
      glBegin(GL_LINES);
      glVertex2i(pr - 16, pt - 16);
      glVertex2i(pr - 4, pt - 4);
      glVertex2i(pr - 4, pt - 16);
      glVertex2i(pr - 16, pt - 4);
      glEnd();
    }
    else  //close object selectable
    {
      glLoadName((wn->name * 10000) + GLWIN_CLOSE);
      glRecti(pr - 16, pt - 16, pr - 4, pt - 4);
    }
  }
  if (wn->resize)  //resize object
  {
    if (mode == GL_RENDER)
    {
      glColor3ub(0, 0, 255);
      glBegin(GL_LINES);
      glVertex2i(pr - 10, pb);
      glVertex2i(pr, pb + 10);
      glVertex2i(pr - 7, pb);
      glVertex2i(pr, pb + 7);
      glVertex2i(pr - 4, pb);
      glVertex2i(pr, pb + 4);
      glEnd();
    }
    else  //resize object selectable
    {
      glLoadName((wn->name * 10000) + GLWIN_RESIZE);
      glBegin(GL_TRIANGLES);
      glVertex2i(pr - 10, pb);
      glVertex2i(pr, pb);
      glVertex2i(pr, pb + 10);
      glEnd();
    }
  }
}

//draw button object
void MyGLWin::DrawButton(GLenum mode, glbt_obj *bt)
{
  int pl = bt->pwn->left + bt->left, pt = glutGet(GLUT_WINDOW_HEIGHT) - bt->pwn->top - bt->top - 20;
  if (mode == GL_RENDER)
  {
    glColor3ub(0, 0, 255);
    glBegin(GL_LINE_LOOP);
    glVertex2i(pl, pt - 30);
    glVertex2i(pl + bt->width, pt - 30);
    glVertex2i(pl + bt->width, pt);
    glVertex2i(pl, pt);
    glEnd();
    if (bt->defalt)
    {
      glBegin(GL_LINES);
      glVertex2i(pl, pt - 29);
      glVertex2i(pl + bt->width, pt - 29);
      glVertex2i(pl, pt - 1);
      glVertex2i(pl + bt->width, pt - 1);
      glEnd();
    }
    glColor3ub(255, 255, 255);
    glRasterPos2i(pl + 10, pt - 18);
    glutBitmapString(GLUT_BITMAP_HELVETICA_10, (const unsigned char *)bt->text);
  }
  else  //button object selectable
  {
    glLoadName((bt->pwn->name * 10000) + bt->value);
    glRecti(pl, pt - 30, pl + bt->width, pt);
  }
}

//draw input object
bool MyGLWin::DrawInput(GLenum mode, glin_obj *in)
{
  int pl = in->pwn->left + in->left, pt = glutGet(GLUT_WINDOW_HEIGHT) - in->pwn->top - in->top - 20, pr = pl + in->width;
  if (mode == GL_RENDER)
  {
    glColor3ub(0, 0, 255);
    glBegin(GL_LINE_LOOP);
    glVertex2i(pl, pt - 20);
    glVertex2i(pr, pt - 20);
    glVertex2i(pr, pt);
    glVertex2i(pl, pt);
    glEnd();
    glColor3ub(255, 255, 255);
    glRasterPos2i(pl + 5, pt - 13);
    unsigned int cnt = 0;
    char *ch = in->text;
    ch += in->first;
    GLint rpos[4] = {0, 0, 0, 0};
    while (*ch && (rpos[0] < (pr - 14)))
    {
      if ((in->name == currInput) && (cnt++ == (in->cursor - in->first))) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, '|');
      glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *ch);
      ch++;
      glGetIntegerv(GL_CURRENT_RASTER_POSITION, rpos);
    }
    if ((in->name == currInput) && (cnt == (in->cursor - in->first))) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, '|');
    if (*ch && (cnt < (in->cursor - in->first)))
    {
      in->first++;  //move first char to show
      return true;
    }
  }
  else  //input object selectable to gain focus (for keys)
  {
    glLoadName((in->pwn->name * 10000) + (in->name * 100) + GLWIN_TEXT);
    glRecti(pl, pt - 20, pr, pt);
  }
  return false;
}

//draw label object
void MyGLWin::DrawLabel(gllb_obj *lb)
{
  glColor3ub(255, 255, 255);
  glRasterPos2i(lb->pwn->left + lb->left, glutGet(GLUT_WINDOW_HEIGHT) - lb->pwn->top - lb->top - 28);
  glutBitmapString(GLUT_BITMAP_HELVETICA_10, (const unsigned char *)lb->text);
}

//draw list object
void MyGLWin::DrawList(GLenum mode, glls_obj *ls)
{
  int pl = ls->pwn->left + ls->left, pt = glutGet(GLUT_WINDOW_HEIGHT) - ls->pwn->top - ls->top - 20
    , pr = pl + ls->width, pb = pt - 100;
  if (mode == GL_RENDER)
  {
    glColor3ub(0, 0, 255);
    glBegin(GL_LINE_LOOP);
    glVertex2i(pl, pb);
    glVertex2i(pr, pb);
    glVertex2i(pr, pt);
    glVertex2i(pl, pt);
    glEnd();
    FILE *text;
    if ((text = fopen(ls->fname, "r")))
    {
      int cnt;
      char chr;
      if (ls->line)
      {
        for (cnt = 0; cnt < ls->line; cnt++)
        {
          while (((chr = getc(text)) != EOF) && (chr != '\n'));  //nothing
        }
        glBegin(GL_LINE_STRIP);  //display up arrow scroll object
        glVertex2i(pr - 12, pt);
        glVertex2i(pr - 12, pt - 12);
        glVertex2i(pr - 6, pt - 1);
        glVertex2i(pr, pt - 12);
        glVertex2i(pr - 12, pt - 12);
        glEnd();
      }
      char *end, *ch;
      int py = pt - 13;
      GLint rpos[4];
      glColor3ub(255, 255, 255);
      for (cnt = 0; cnt < 5; cnt++)
      {
        if (fgets(ls->items[cnt], 256, text))
        {
          if ((end = strchr(ls->items[cnt], '\n'))) *end = '\0';  //remove \n
          else while (((chr = getc(text)) != EOF) && (chr != '\n'));  //nothing
          glRasterPos2i(pl + 5, py);
          ch = ls->items[cnt];
          rpos[0] = 0;
          while (*ch && (rpos[0] < (pr - 32)))
          {
            glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, *ch);
            ch++;
            glGetIntegerv(GL_CURRENT_RASTER_POSITION, rpos);
          }
          if (*ch) glutBitmapString(GLUT_BITMAP_HELVETICA_10, MTF);
          py -= 20;
        }
        else break;
      }
      if ((ls->more = ((chr = getc(text)) != EOF)))
      {
        glColor3ub(0, 0, 255);
        glBegin(GL_LINE_STRIP);  //display down arrow scroll object
        glVertex2i(pr - 12, pb);
        glVertex2i(pr - 12, pb + 12);
        glVertex2i(pr - 6, pb + 1);
        glVertex2i(pr, pb + 12);
        glVertex2i(pr - 12, pb + 12);
        glEnd();
      }
      fclose(text);
    }
  }
  else  //list items/scroll up/down objects selectable, also give object focus (for mouse wheel)
  {
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_LIST1);
    glRecti(pl, pt - 19, pr - 16, pt - 1);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_LIST2);
    glRecti(pl, pt - 39, pr - 16, pt - 21);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_LIST3);
    glRecti(pl, pt - 59, pr - 16, pt - 41);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_LIST4);
    glRecti(pl, pt - 79, pr - 16, pt - 61);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_LIST5);
    glRecti(pl, pt - 99, pr - 16, pt - 81);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_UP);
    glRecti(pr - 12, pt - 12, pr, pt);
    glLoadName((ls->pwn->name * 10000) + (ls->name * 100) + GLWIN_DOWN);
    glRecti(pr - 12, pb, pr, pb + 12);
  }
}

//draw view object
void MyGLWin::DrawView(GLenum mode, glvw_obj *vw)
{
  int pl = vw->pwn->left + vw->left, pt = glutGet(GLUT_WINDOW_HEIGHT) - vw->pwn->top - vw->top - 20
    , pr = vw->pwn->left + vw->pwn->width - vw->right, pb = glutGet(GLUT_WINDOW_HEIGHT) - vw->pwn->top - vw->pwn->height + vw->bottom;
  if (mode == GL_RENDER)
  {
    glColor3ub(0, 0, 255);
    glBegin(GL_LINE_LOOP);
    glVertex2i(pl, pb);
    glVertex2i(pr, pb);
    glVertex2i(pr, pt);
    glVertex2i(pl, pt);
    glEnd();
    FILE *text;
    if ((text = fopen(vw->fname, "r")))
    {
      char chr;
      if (vw->line)
      {
        for (int cnt = 0; cnt < vw->line; cnt++)
        {
          while (((chr = getc(text)) != EOF) && (chr != '\n'));  //nothing
        }
        glBegin(GL_LINE_STRIP);  //display up arrow scroll object
        glVertex2i(pr - 12, pt);
        glVertex2i(pr - 12, pt - 12);
        glVertex2i(pr - 6, pt - 1);
        glVertex2i(pr, pt - 12);
        glVertex2i(pr - 12, pt - 12);
        glEnd();
      }
      int py = pt - 18;
      GLint rpos[4];
      glColor3ub(255, 255, 255);
      glRasterPos2i(pl + 10, py);
      chr = getc(text);
      while ((vw->more = (chr != EOF)))
      {
        if (chr == '\n')  //CR
        {
          if ((chr = getc(text)) != EOF)
          {
            py -= 14;
            if (py >= (pb + 10)) glRasterPos2i(pl + 10, py);
            else
            {
              glColor3ub(0, 0, 255);
              glBegin(GL_LINE_STRIP);  //display down arrow scroll object
              glVertex2i(pr - 12, pb);
              glVertex2i(pr - 12, pb + 12);
              glVertex2i(pr - 6, pb + 1);
              glVertex2i(pr, pb + 12);
              glVertex2i(pr - 12, pb + 12);
              glEnd();
              break;
            }
          }
        }
        else
        {
          if (chr == '\t') glRasterPos2i(pl + 10 + vw->tab, py);  //tab
          else
          {
            glGetIntegerv(GL_CURRENT_RASTER_POSITION, rpos);
            if (rpos[0] < (pr - 22)) glutBitmapCharacter(GLUT_BITMAP_HELVETICA_10, chr);
          }
          chr = getc(text);
        }
      }
      fclose(text);
    }
  }
  else  //scroll up/down objects selectable, also give object focus (for mouse wheel)
  {
    glLoadName((vw->pwn->name * 10000) + (vw->name * 100) + GLWIN_UP);
    glRecti(pr - 12, pt - 12, pr, pt);
    glLoadName((vw->pwn->name * 10000) + (vw->name * 100) + GLWIN_DOWN);
    glRecti(pr - 12, pb, pr, pb + 12);
  }
}

//make next/prev input object currInput
void MyGLWin::TabInput(bool nx)
{
  glin_obj *in;
  unsigned char *tp;
  (nx ? GLWinLL.Start(1) : GLWinLL.End(1));
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_INPUT)
    {
      in = (glin_obj *)tp;
      if (in->name == currInput) break;
    }
    (nx ? GLWinLL.Next(1) : GLWinLL.Prev(1));
  }
  (nx ? GLWinLL.Next(1) : GLWinLL.Prev(1));
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_INPUT)
    {
      in = (glin_obj *)tp;
      if (in->pwn->name == (selection / 10000))
      {
        currInput = in->name;
        return;
      }
    }
    (nx ? GLWinLL.Next(1) : GLWinLL.Prev(1));
  }
  (nx ? GLWinLL.Start(1) : GLWinLL.End(1));
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_INPUT)
    {
      in = (glin_obj *)tp;
      if (in->pwn->name == (selection / 10000))
      {
        currInput = in->name;
        break;
      }
    }
    (nx ? GLWinLL.Next(1) : GLWinLL.Prev(1));
  }
}

//destroy child objects for parent window (pw), or all objects if passed null
void MyGLWin::GLWinDestroy(glwn_obj *pw)
{
  glwn_obj *wn;
  glbt_obj *bt;
  glin_obj *in;
  gllb_obj *lb;
  glls_obj *ls;
  glvw_obj *vw;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    switch (*tp)
    {
      case GLWIN_WINDOW:
        if (!pw)
        {
          wn = (glwn_obj *)tp;
          delete wn;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
      case GLWIN_BUTTON:
        bt = (glbt_obj *)tp;
        if (!pw || (bt->pwn == pw))
        {
          delete bt;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
      case GLWIN_INPUT:
        in = (glin_obj *)tp;
        if (!pw || (in->pwn == pw))
        {
          delete in;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
      case GLWIN_LABEL:
        lb = (gllb_obj *)tp;
        if (!pw || (lb->pwn == pw))
        {
          delete lb;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
      case GLWIN_LIST:
        ls = (glls_obj *)tp;
        if (!pw || (ls->pwn == pw))
        {
          delete ls;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
      case GLWIN_VIEW:
        vw = (glvw_obj *)tp;
        if (!pw || (vw->pwn == pw))
        {
          delete vw;
          GLWinLL.Delete(1);
        }
        else GLWinLL.Next(1);
        break;
    }
  }
}

//2D GUI active
bool MyGLWin::On()
{
  return GLWinLL.Num();
}

//create window object
void MyGLWin::CreateWin(int left, int top, int width, int height, const char *title, bool close, bool resize)
{
  glwn_obj glwn;
  glwn.type = GLWIN_WINDOW;
  glwn.close = close;  //display close object
  glwn.resize = resize;  //display resize object
  glwn.name = names++;
  glwn.psel = selection;  //previous selection
  int gw = glutGet(GLUT_WINDOW_WIDTH), gh = glutGet(GLUT_WINDOW_HEIGHT);
  glwn.minW = width;  //minimum width
  glwn.width = ((left < -1) && ((gw - 20) >= width) ? gw - 20 : width);
  glwn.minH = height;  //minimum height
  glwn.height = ((top < -1) && ((gh - 20) >= height) ? gh - 20 : height);
  if (left > -1) glwn.left = left;
  else if ((gw - glwn.width) > 0) glwn.left = (gw - glwn.width) / 2;
  else glwn.left = 0;
  if (top > -1) glwn.top = top;
  else if ((gh - glwn.height) > 0) glwn.top = (gh - glwn.height) / 2;
  else glwn.top = 0;
  strcpy(glwn.title, title);  //title bar text
  selection = glwn.name * 10000;
  lastWin = (glwn_obj *)GLWinLL.Write(new glwn_obj(glwn));
}

//create button object
void MyGLWin::AddButton(int left, int top, int width, const char *text, bool defalt, int value)
{
  if (lastWin)  //check parent window object exists
  {
    glbt_obj glbt = {GLWIN_BUTTON, lastWin, defalt, left, top, width, value};
    strcpy(glbt.text, text);
    GLWinLL.Write(new glbt_obj(glbt));
  }
}

//create input object, return name of input object
int MyGLWin::AddInput(int left, int top, int width, const char *text, bool lower, int max)
{
  if (!lastWin) return -1;  //check parent window object exists
  glin_obj glin = {GLWIN_INPUT, lastWin, lower, names++, left, top, width, strlen(text), 0, max};
  strcpy(glin.text, text);  //default text
  currInput = glin.name;  //set input object focus (for keys)
  lastInput = (glin_obj *)GLWinLL.Write(new glin_obj(glin));
  return glin.name;
}

//create label object, return name of label object
int MyGLWin::AddLabel(int left, int top, const char *text)
{
  if (!lastWin) return -1;  //check parent window object exists
  gllb_obj gllb = {GLWIN_LABEL, lastWin, names++, left, top};
  strcpy(gllb.text, text);
  GLWinLL.Write(new gllb_obj(gllb));
  return gllb.name;
}

//create list object
void MyGLWin::AddList(int left, int top, int width, const char *fl)
{
  if (lastWin && lastInput)  //check parent window/linked input object exists
  {
    glls_obj glls = {GLWIN_LIST, lastWin, lastInput, false, names++, left, top, width, 0, {"", "", "", "", ""}};
    strcpy(glls.fname, fl);  //file to list
    currScroll = (glls.pwn->name * 10000) + (glls.name * 100);  //set scroll object focus (for mouse wheel)
    GLWinLL.Write(new glls_obj(glls));
  }
}

//create view object, return name of view object
int MyGLWin::AddView(int left, int top, int right, int bottom, int tab, const char *fl)
{
  if (!lastWin) return -1;  //check parent window object exists
  glvw_obj glvw = {GLWIN_VIEW, lastWin, false, names++, left, top, right, bottom, tab, 0};
  strcpy(glvw.fname, fl);  //file to view
  currScroll = (glvw.pwn->name * 10000) + (glvw.name * 100);  //set scroll object focus (for mouse wheel)
  GLWinLL.Write(new glvw_obj(glvw));
  return glvw.name;
}

//draw all 2D GUI objects
bool MyGLWin::Draw(GLenum mode)
{
  bool result = false;
  glwn_obj *wn;
  glbt_obj *bt;
  glin_obj *in;
  gllb_obj *lb;
  glls_obj *ls;
  glvw_obj *vw;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    switch (*tp)
    {
      case GLWIN_WINDOW:
        wn = (glwn_obj *)tp;
        DrawWin(mode, wn);
        break;
      case GLWIN_BUTTON:
        bt = (glbt_obj *)tp;
        DrawButton(mode, bt);
        break;
      case GLWIN_INPUT:
        in = (glin_obj *)tp;
        if (DrawInput(mode, in)) result = true;
        break;
      case GLWIN_LABEL:
        lb = (gllb_obj *)tp;
        DrawLabel(lb);
        break;
      case GLWIN_LIST:
        ls = (glls_obj *)tp;
        DrawList(mode, ls);
        break;
      case GLWIN_VIEW:
        vw = (glvw_obj *)tp;
        DrawView(mode, vw);
        break;
    }
    GLWinLL.Next(1);
  }
  return result;
}

//return input object text
char *MyGLWin::GetInput(int name)
{
  int gn = (name ? name : currInput);
  glin_obj *in;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_INPUT)
    {
      in = (glin_obj *)tp;
      if (in->name == gn) return in->text;
    }
    GLWinLL.Next(1);
  }
  return 0;
}

//change input object text
void MyGLWin::PutInput(int name, const char *text)
{
  int gn = (name ? name : currInput);
  glin_obj *in;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_INPUT)
    {
      in = (glin_obj *)tp;
      if (in->name == gn)
      {
        strncpy(in->text, text, in->max);
        if (strlen(text) >= in->max) in->text[in->max] = '\0';
        in->cursor = strlen(in->text);
        in->first = 0;
        break;
      }
    }
    GLWinLL.Next(1);
  }
}

//change label object text
void MyGLWin::PutLabel(int name, const char *text)
{
  gllb_obj *lb;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_LABEL)
    {
      lb = (gllb_obj *)tp;
      if (lb->name == name)
      {
        strcpy(lb->text, text);
        break;
      }
    }
    GLWinLL.Next(1);
  }
}

//reset view object first line shown to 0
void MyGLWin::ResetView(int name)
{
  glvw_obj *vw;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_VIEW)
    {
      vw = (glvw_obj *)tp;
      if (vw->name == name)
      {
        vw->line = 0;
        currScroll = (vw->pwn->name * 10000) + (vw->name * 100);  //set scroll object focus (for mouse wheel)
        break;
      }
    }
    GLWinLL.Next(1);
  }
}

//mouse left-click and drag motion
void MyGLWin::Move(int x, int y)
{
  int gw, gh;
  glwn_obj *wn;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_WINDOW)
    {
      wn = (glwn_obj *)tp;
      if (wn->name == (selection / 10000))
      {
        gw = glutGet(GLUT_WINDOW_WIDTH);
        gh = glutGet(GLUT_WINDOW_HEIGHT);
        switch (selection % 100)
        {
          case GLWIN_RESIZE:  //resize window object
            wn->width += x;
            if (wn->width < wn->minW) wn->width = wn->minW;
            else if (wn->width > (gw - wn->left)) wn->width = gw - wn->left;
            wn->height += y;
            if (wn->height < wn->minH) wn->height = wn->minH;
            else if (wn->height > (gh - wn->top)) wn->height = gh - wn->top;
            break;
          case GLWIN_TITLE:  //move window object
            wn->left += x;
            if (wn->left < 0) wn->left = 0;
            else if (wn->left > (gw - 20)) wn->left = gw - 20;
            wn->top += y;
            if (wn->top < 0) wn->top = 0;
            else if (wn->top > (gh - 20)) wn->top = gh - 20;
            break;
        }
        break;
      }
    }
    GLWinLL.Next(1);
  }
}

//scroll object/mouse wheel action
void MyGLWin::Scroll(int gs)
{
  if (gs) selection = currScroll + gs;
  glls_obj *ls;
  glvw_obj *vw;
  unsigned char *tp;
  GLWinLL.Start(1);
  while ((tp = (unsigned char *)GLWinLL.Read(1)))
  {
    if (*tp == GLWIN_LIST)
    {
      ls = (glls_obj *)tp;
      if (ls->name == ((selection % 10000) / 100))
      {
        switch (selection % 100)
        {
          case GLWIN_UP: if (ls->line) ls->line--; break;
          case GLWIN_DOWN:
            if (ls->more)  //more list items to follow
            {
              ls->line++;
              ls->more = false;
            }
            break;
        }
        currScroll = (ls->pwn->name * 10000) + (ls->name * 100);  //set scroll object focus (for mouse wheel)
        break;
      }
    }
    else if (*tp == GLWIN_VIEW)
    {
      vw = (glvw_obj *)tp;
      if (vw->name == ((selection % 10000) / 100))
      {
        switch (selection % 100)
        {
          case GLWIN_UP: if (vw->line) vw->line--; break;
          case GLWIN_DOWN:
            if (vw->more)  //more text to follow
            {
              vw->line++;
              vw->more = false;
            }
            break;
        }
        currScroll = (vw->pwn->name * 10000) + (vw->name * 100);  //set scroll object focus (for mouse wheel)
        break;
      }
    }
    GLWinLL.Next(1);
  }
}

//close window object and child objects, or all objects
void MyGLWin::Close(bool all)
{
  if (all) GLWinDestroy(0);
  else
  {
    glwn_obj *wn;
    unsigned char *tp;
    GLWinLL.Start(2);
    while ((tp = (unsigned char *)GLWinLL.Read(2)))
    {
      if (*tp == GLWIN_WINDOW)
      {
        wn = (glwn_obj *)tp;
        if (wn->name == (selection / 10000))
        {
          selection = wn->psel;  //previous selection
          GLWinDestroy(wn);
          delete wn;
          GLWinLL.Delete(2);
          break;
        }
      }
      GLWinLL.Next(2);
    }
  }
  if (!On())
  {
    Reset();
    glutAttachMenu(GLUT_RIGHT_BUTTON);
  }
}

//mouse left-click object selection
int MyGLWin::Select(int state, int x, int y)
{
  GLint hits, vpt[4];
  GLuint selbuf[4];
  glGetIntegerv(GL_VIEWPORT, vpt);
  glSelectBuffer(4, selbuf);
  glRenderMode(GL_SELECT);
  glInitNames();
  glPushName(0);
  glDisable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluPickMatrix(x, vpt[3] - y, 1.0, 1.0, vpt);
  gluOrtho2D(0.0, glutGet(GLUT_WINDOW_WIDTH), 0.0, glutGet(GLUT_WINDOW_HEIGHT));
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  Draw(GL_SELECT);
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();
  glEnable(GL_DEPTH_TEST);
  if (!(hits = glRenderMode(GL_RENDER))) return 0;
  selection = selbuf[3];
  int sel = selection % 100;
  if (state == GLUT_UP)
  {
    switch (sel)
    {
      case GLWIN_TEXT: currInput = (selection % 10000) / 100; break;  //set input object focus (for keys)
      case GLWIN_UP: case GLWIN_DOWN: Scroll(0); break;
      case GLWIN_LIST1: case GLWIN_LIST2: case GLWIN_LIST3: case GLWIN_LIST4: case GLWIN_LIST5:
        glls_obj *ls;
        unsigned char *tp;
        GLWinLL.Start(1);
        while ((tp = (unsigned char *)GLWinLL.Read(1)))
        {
          if (*tp == GLWIN_LIST)
          {
            ls = (glls_obj *)tp;
            if (ls->name == ((selection % 10000) / 100))
            {
              strcpy(ls->lin->text, ls->items[sel - GLWIN_LIST1]);  //set linked input object text
              ls->lin->cursor = strlen(ls->lin->text);
              ls->lin->first = 0;
              currScroll = (ls->pwn->name * 10000) + (ls->name * 100);  //set scroll object focus (for mouse wheel)
              break;
            }
          }
          GLWinLL.Next(1);
        }
        break;
    }
  }
  return sel;
}

//keyboard input
void MyGLWin::Key(unsigned char key)
{
  if (!currInput) return;
#ifdef __MINGW32__
  if (key == 9) (glutGetModifiers() != GLUT_ACTIVE_SHIFT ? TabInput(true) : TabInput(false));  //Tab
#else
  if (key == 9) (glutGetModifiers() != GLUT_ACTIVE_CTRL ? TabInput(true) : TabInput(false));  //Tab
#endif
  else
  {
    unsigned int cnt;
    glin_obj *in;
    unsigned char *tp;
    GLWinLL.Start(1);
    while ((tp = (unsigned char *)GLWinLL.Read(1)))
    {
      if (*tp == GLWIN_INPUT)
      {
        in = (glin_obj *)tp;
        if (in->name == currInput)  //input object with focus
        {
          switch (key)
          {
            case 8:  //BS
              if (in->cursor)
              {
                for (cnt = in->cursor - 1; cnt < strlen(in->text); cnt++) in->text[cnt] = in->text[cnt + 1];
                if (in->cursor) in->cursor--;
                if (in->first) in->first--;
              }
              break;
            case 28:  //LA
              if (in->cursor) in->cursor--;
              if (in->first > in->cursor) in->first = in->cursor;
              break;
            case 29: if (in->cursor < strlen(in->text)) in->cursor++; break;  //RA
            default:
              if ((key >= 32) && (key <= 126) && (strlen(in->text) < in->max))
              {
                for (cnt = strlen(in->text) + 1; cnt > in->cursor; cnt--) in->text[cnt] = in->text[cnt - 1];
                in->text[in->cursor] = (in->lower ? tolower(key) : key);
                in->cursor++;
              }
          }
          break;
        }
      }
      GLWinLL.Next(1);
    }
  }
}

MyGLWin::~MyGLWin()
{
  GLWinDestroy(0);
}
